#
# This file was generated from the asynchronous client at invenio/transfer/__init__.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Different transfers to the invenio repository."""

from .base import Transfer
from .registry import transfer_registry

__all__ = ("Transfer", "transfer_registry")

