#
# This file was generated from the asynchronous client at invenio/__init__.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


from .client import SyncInvenioRepositoryClient

__all__ = ("SyncInvenioRepositoryClient", )
